package org.processmining.miminer;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.processmining.framework.plugin.PluginContext;
import org.processmining.miminer.models.ArtifactRelationModel;
import org.processmining.miminer.models.ArtifactTypeModel;
import org.processmining.miminer.models.ArtifactTypeModel.ArtifactState;
import org.processmining.miminer.models.ArtifactTypeModel.ArtifactTransition;
import org.processmining.miminer.models.StateEntry;

import com.google.common.collect.HashMultiset;
import com.google.common.collect.ListMultimap;
import com.google.common.collect.MultimapBuilder;
import com.google.common.collect.Multiset;
import com.google.common.collect.SetMultimap;

public class MIMiner {

	public static final String initialState = "[Initial State]";
	public static final String finalState = "[Final State]";
	public static final long maxTransitionDuration = 500;  
	
	public static Object[] mine(PluginContext context, MIMinerConfiguration configuration) {
		MIMinerResults results = new MIMinerResults(configuration);
		context.getProgress().setMaximum(configuration.instances.size());
		
		// Create one model for each artifact type
		for (String type : configuration.types) {
			results.artifactTypeModels.put(type, new ArtifactTypeModel(type));
		}
		
		// Update the models for each artifact instance
		for (int i = 0; i < configuration.instances.size(); i++) {
			ArtifactTypeModel model = results.artifactTypeModels.get(configuration.instanceTypeMap.get(i));
			addInstance(configuration, model, i, results.artifactRelationModel, results);
			context.getProgress().inc();
		}
		
		return new Object[] {
				results,
				configuration.log
		};
	}

	private static void addInstance(MIMinerConfiguration configuration, ArtifactTypeModel model, int instance,
			ArtifactRelationModel interactionModel, MIMinerResults results) {
		List<StateEntry> stateEntries = new ArrayList<>(configuration.instanceStatesMap.get(instance));
		Set<Integer> relatedInstances = new HashSet<>(configuration.instanceRelationsMap.get(instance));
		Multiset<String> relatedTypes = HashMultiset.create();
		
		// Create a single time-ordered sequence of all state entries of this instance or related to it
		for (Integer relatedInstance : configuration.instanceRelationsMap.get(instance)) {
			// TODO: add configuration of relation time window
			if (relatedInstance == instance) continue;
			stateEntries.addAll(configuration.instanceStatesMap.get(relatedInstance));
			relatedTypes.add(configuration.instanceTypeMap.get(relatedInstance));
		}
		for (StateEntry entry : configuration.instanceRelatedStatesMap.get(instance)) {
			if (configuration.instanceRelationsMap.get(instance).contains(entry.artifactInstance)) {
				// All these state entries have already been added above
				continue;
			}
			else {
				stateEntries.add(entry);
				
				if (!relatedInstances.contains(entry.artifactInstance)) {
					relatedTypes.add(configuration.instanceTypeMap.get(entry.artifactInstance));
					relatedInstances.add(entry.artifactInstance);
				}
			}
		}
		stateEntries.sort(new MIMiner.StateEntryComparator());
		relatedInstances.remove(instance);
		
		// Determine the related types and their frequencies
		interactionModel.addRelations(model.artifactType, relatedTypes);
		
		// Initialise the states of all relevant instances
		Map<Integer, String> instanceStates = new HashMap<>();
		instanceStates.put(instance, initialState);
		for (Integer relatedInstance : relatedInstances) {
			instanceStates.put(relatedInstance, initialState);
		}
		
		// Go through the list of state entries to group the overlapping entries as transitions, removing self-loops and double transitions
		ListMultimap<Long, StateEntry> groupedEntries = MultimapBuilder.treeKeys().arrayListValues().build();
		SetMultimap<Long, Integer> groupedInstances = MultimapBuilder.treeKeys().hashSetValues().build();
		Long currentTimestamp = null;
		for (StateEntry entry : stateEntries) {
			String stateName = getStateName(configuration, entry);
			String previousStateName = instanceStates.get(entry.artifactInstance);
			
			if (!stateName.equals(previousStateName)) {
				if (!groupedInstances.get(entry.getTimestamp().getTime()).contains(entry.artifactInstance)) {
					if (instance == entry.artifactInstance.intValue()) {
						// stateEntry of current artifact type
						currentTimestamp = entry.getTimestamp().getTime();
						groupedEntries.put(currentTimestamp, entry);
					}
					else {
						// related stateEntry
						if (currentTimestamp != null && groupedInstances.get(currentTimestamp).contains(entry.artifactInstance)) {
							// Forced end of the transition duration
							currentTimestamp = null;
						}
						
						if (currentTimestamp == null || entry.getTimestamp().getTime() - currentTimestamp > maxTransitionDuration) {
							groupedEntries.put(entry.getTimestamp().getTime(), entry);
						}
						else {
							// First related entry of this type within the transition duration window after an instance transition
							groupedEntries.put(currentTimestamp, entry);
							groupedInstances.put(currentTimestamp, entry.artifactInstance);
						}
					}
					
					instanceStates.put(entry.artifactInstance, stateName);
					groupedInstances.put(entry.getTimestamp().getTime(), entry.artifactInstance);
				}
				else {
					// Multiple entries for this instance with the same timestamp
					continue;
				}
			}
			else {
				// Self-loop
				continue;
			}
		}
		
//		System.out.println(groupedEntries);
		
		// Reinitialise the states of all relevant instances
		instanceStates.clear();
		instanceStates.put(instance, initialState);
		for (Integer relatedInstance : relatedInstances) {
			instanceStates.put(relatedInstance, initialState);
		}
		
		// Determine the state of each relevant instance at all points in time and create the model
		ArtifactState modelState = model.addState(initialState);
		ArtifactTransition modelTransition = null;
		Map<Integer, Long> instanceLastUpdated = new HashMap<>();
		Map<String, Long> typeLastUpdated = new HashMap<>();
		Long lastUpdated = 0L;
		
		for (Long timestamp: groupedEntries.keySet()) {
			// Update the current state and then update the model
			if (groupedInstances.get(timestamp).contains(instance)) {
				// First entry is a transition in the model
				if (instanceLastUpdated.get(instance) == null) {
					// Set initial time for all active tokens
					for (Integer anInstance : instanceStates.keySet()) {
						if (!instanceStates.get(anInstance).equals(initialState)) {
							instanceLastUpdated.put(anInstance, timestamp);
							typeLastUpdated.put(configuration.instanceTypeMap.get(anInstance), timestamp);
						}
					}
				}
				else {
					// Update state statistics
					Multiset<ArtifactState> marking = HashMultiset.create();
					Map<String,Multiset<ArtifactState>> typeMarkings = new HashMap<>();
					marking.add(modelState);
					Multiset<ArtifactState> typeMarking = typeMarkings.get(configuration.instanceTypeMap.get(instance));
					if (typeMarking == null) {
						typeMarking = HashMultiset.create();
						typeMarkings.put(configuration.instanceTypeMap.get(instance), typeMarking);
					}
					typeMarking.add(modelState);
					
					for (Integer relatedInstance : relatedInstances) {
						if (!instanceStates.get(relatedInstance).equals(initialState)) {
							ArtifactState relatedState = model.addRelatedStatePeriod(modelState, configuration.instanceTypeMap.get(
									relatedInstance), instanceStates.get(relatedInstance), timestamp - instanceLastUpdated.get(
											relatedInstance));
							marking.add(relatedState);
							typeMarking = typeMarkings.get(configuration.instanceTypeMap.get(relatedInstance));
							if (typeMarking == null) {
								typeMarking = HashMultiset.create();
								typeMarkings.put(configuration.instanceTypeMap.get(relatedInstance), typeMarking);
							}
							typeMarking.add(relatedState);
						}
					}
					
					model.addRelatedMarking(modelState, marking, timestamp - lastUpdated);
					for (String type : typeMarkings.keySet()) {
						model.addRelatedTypeMarkings(type, modelState, typeMarkings.get(type), timestamp - typeLastUpdated.get(type));
					}
				}
				
				// Update transition statistics
				Map<Integer, ArtifactTransition> relatedTransitions = new HashMap<>();
				ArtifactState newModelState = null;
				for (StateEntry entry : groupedEntries.get(timestamp)) {
					if (entry.artifactInstance.intValue() == instance) {
						// Create model transition and related statistics
						newModelState = model.addState(getStateName(configuration, entry));
						modelTransition = model.addTransition(instanceStates.get(entry.artifactInstance), getStateName(
								configuration, entry));
						
						if (instanceLastUpdated.get(instance) != null) {
							model.addStatePeriod(modelState.stateName, timestamp - instanceLastUpdated.get(instance));
						}
					}
					else {
						// Create related transitions and statistics
						ArtifactTransition relatedTransition = model.addTransitionPostTransitionCount(modelTransition,
								configuration.instanceTypeMap.get(entry.artifactInstance), instanceStates.get(entry.artifactInstance),
								getStateName(configuration, entry));
						relatedTransitions.put(entry.artifactInstance,relatedTransition);
					}
					
					instanceStates.put(entry.artifactInstance, getStateName(configuration, entry));
					instanceLastUpdated.put(entry.artifactInstance, timestamp);
					typeLastUpdated.put(configuration.instanceTypeMap.get(entry.artifactInstance), timestamp);
				}
				
				// Update transition total marking statistics
				Multiset<ArtifactTransition> multiset = HashMultiset.create();
				Multiset<ArtifactTransition> marking = HashMultiset.create();
				multiset.add(modelTransition);
				marking.add(modelTransition);
				
				for (Integer relatedInstance : relatedInstances) {
					if (!relatedTransitions.containsKey(relatedInstance)) {
						if (!instanceStates.get(relatedInstance).equals(initialState)) {
							marking.add(model.addTransitionPostStateCount(modelTransition,
									configuration.instanceTypeMap.get(relatedInstance), instanceStates.get(relatedInstance)));
							instanceLastUpdated.put(relatedInstance, timestamp);
							typeLastUpdated.put(configuration.instanceTypeMap.get(relatedInstance), timestamp);
						}
					}
					else {
						multiset.add(relatedTransitions.get(relatedInstance));
						marking.add(relatedTransitions.get(relatedInstance));
					}
				}
				model.addTransitionPostMultiset(modelTransition, multiset);
				model.addTransitionPostMarking(modelTransition, marking);
				
				modelState = newModelState;
			}
			else {
				// Only related entries
				if (instanceLastUpdated.get(instance) == null) {
					// Related entries before the instance has started
					for (StateEntry entry : groupedEntries.get(timestamp)) {
						instanceStates.put(entry.artifactInstance, getStateName(configuration, entry));
					}
				}
				else {
					// Update state statistics
					Multiset<ArtifactState> marking = HashMultiset.create();
					Map<String,Multiset<ArtifactState>> typeMarkings = new HashMap<>();
					marking.add(modelState);
					
					Multiset<ArtifactState> typeMarking = null;
					Set<String> changedTypes = new HashSet<>();
					for (Integer relatedInstance : relatedInstances) {
						if (!instanceStates.get(relatedInstance).equals(initialState)) {
							ArtifactState relatedState = model.addRelatedStatePeriod(modelState, configuration.instanceTypeMap.get(
									relatedInstance), instanceStates.get(relatedInstance), timestamp - instanceLastUpdated.get(
											relatedInstance));
							marking.add(relatedState);
							typeMarking = typeMarkings.get(configuration.instanceTypeMap.get(relatedInstance));
							if (typeMarking == null) {
								typeMarking = HashMultiset.create();
								typeMarkings.put(configuration.instanceTypeMap.get(relatedInstance), typeMarking);
							}
							typeMarking.add(relatedState);
							changedTypes.add(configuration.instanceTypeMap.get(relatedInstance));
						}
					}
					
					if (changedTypes.contains(configuration.instanceTypeMap.get(instance))) {
						typeMarking = typeMarkings.get(configuration.instanceTypeMap.get(instance));
						typeMarking.add(modelState);
					}
					
					model.addRelatedMarking(modelState, marking, timestamp - lastUpdated);
					for (String type : typeMarkings.keySet()) {
						model.addRelatedTypeMarkings(type, modelState, typeMarkings.get(type), timestamp - typeLastUpdated.get(type));
					}
					
					// Update transition statistics
					Set<Integer> relatedTransitionInstances = new HashSet<>();
					Multiset<ArtifactTransition> multiset = HashMultiset.create();
					Multiset<ArtifactTransition> transitionMarking = HashMultiset.create();
					transitionMarking.add(model.transitionFromState(modelState));
					
					for (StateEntry entry : groupedEntries.get(timestamp)) {
						ArtifactTransition relatedTransition = model.addRelatedTransitionCount(modelState,
								configuration.instanceTypeMap.get(entry.artifactInstance), instanceStates.get(entry.artifactInstance),
								getStateName(configuration, entry));
						multiset.add(relatedTransition);
						transitionMarking.add(relatedTransition);
						relatedTransitionInstances.add(entry.artifactInstance);
						
						instanceStates.put(entry.artifactInstance, getStateName(configuration, entry));
						instanceLastUpdated.put(entry.artifactInstance, timestamp);
						typeLastUpdated.put(configuration.instanceTypeMap.get(entry.artifactInstance), timestamp);
					}
					
					// Update transition total marking statistics
					for (Integer relatedInstance : relatedInstances) {
						if (!relatedTransitionInstances.contains(relatedInstance)) {
							if (!instanceStates.get(relatedInstance).equals(initialState)) {
								transitionMarking.add(model.getTransition(configuration.instanceTypeMap.get(relatedInstance),
										instanceStates.get(relatedInstance), instanceStates.get(relatedInstance)));
							}
						}
					}
					
					model.addRelatedTransitionMultisets(modelState, multiset);
					model.addRelatedTransitionMarking(modelState, transitionMarking);
				}
			}
			
			lastUpdated = timestamp;
		}
		
		// TODO: remove test printing
//		if (modelState != null) {
//			System.out.println(modelState);
//			System.out.println(model.relatedTransitionCounts.row(modelState));
//			System.out.println(model.relatedTransitionMultisets.row(modelState));
//			System.out.println(model.relatedTransitionMarkings.row(modelState));
//			System.out.println(model.relatedStatePeriods.row(modelState));
//			System.out.println(model.relatedMarkings.row(modelState));
//			System.out.println(model.relatedTypeMarkings);
//			Table<ArtifactState, Multiset<ArtifactState>, DescriptiveStatistics> table = model.relatedTypeMarkings.get("w");
//			if (table != null) {
//				System.out.println(model.relatedTypeMarkings.get("w").row(modelState));
//			}
//		}
//
//		if (modelTransition != null) {
//			System.out.println(modelTransition);
//			System.out.println(model.transitionPostStateCounts.row(modelTransition));
//			System.out.println(model.transitionPostTransitionCounts.row(modelTransition));
//			System.out.println(model.transitionPostMultisets.row(modelTransition));
//			System.out.println(model.transitionPostMarkings.row(modelTransition));
//		}
		
		// TODO: do update transition statistics for transition to final state?
		model.addTransition(instanceStates.get(instance), finalState);
		model.addState(finalState);
	}
	
	static String getStateName(MIMinerConfiguration configuration, StateEntry entry) {
		return configuration.classifier.getClassIdentity(entry.event);
	}
	
	static class StateEntryComparator implements Comparator<StateEntry> {
		public int compare(StateEntry entry1, StateEntry entry2) {
			return entry1.getTimestamp().compareTo(entry2.getTimestamp());
		}
	}
}
