package org.processmining.miminer.models;

import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.util.HashMap;
import java.util.Map;

import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.JLabel;
import javax.swing.JWindow;
import javax.swing.SwingConstants;

import org.apache.commons.math3.stat.descriptive.DescriptiveStatistics;
import org.processmining.plugins.graphviz.dot.Dot;
import org.processmining.plugins.graphviz.dot.DotEdge;
import org.processmining.plugins.graphviz.dot.DotNode;
import org.processmining.plugins.graphviz.visualisation.DotPanel;

import com.google.common.collect.HashBasedTable;
import com.google.common.collect.HashMultiset;
import com.google.common.collect.Multiset;
import com.google.common.collect.Table;

public class ArtifactRelationModel {

	public Multiset<String> artifactTypes = HashMultiset.create();
	public Table<String, String, DescriptiveStatistics> relatedTypeCounts = HashBasedTable.create();
	
	private JWindow currentStatisticsWindow;
	private long enterTime;
	private boolean exited;
	
	public ArtifactRelationModel() {
		
	}
	
	public void addRelations(String artifactType, Multiset<String> relatedTypes) {
		artifactTypes.add(artifactType);
		
		for (String relatedType : relatedTypes.elementSet()) {
			DescriptiveStatistics relationCount = relatedTypeCounts.get(artifactType, relatedType);
			if (relationCount == null) {
				relationCount = new DescriptiveStatistics();
				relatedTypeCounts.put(artifactType, relatedType, relationCount);
			}
			
			relationCount.addValue(relatedTypes.count(relatedType));
		}
	}

	public DotPanel createGraph() {
		Dot dot = new Dot();
		DotPanel graph = new DotPanel(dot);
		graph.setBorder(BorderFactory.createEmptyBorder(2, 2, 2, 2));
		graph.setBackground(Color.white);
		
		// Populate graph
		Map<String,DotNode> nodeMap = new HashMap<>();
		for (String type : artifactTypes.elementSet()) {
			DotNode node = dot.addNode(type + "\n" + artifactTypes.count(type));
			nodeMap.put(type, node);

			// TODO: add selection listener
//			node.addSelectionListener(new NodeSelectionlistener(state));
//			node.setSelectable(true);
		}
		for (String fromNode : relatedTypeCounts.rowKeySet()) {
			for (String toNode : relatedTypeCounts.columnKeySet()) {
				if (relatedTypeCounts.contains(fromNode, toNode)) {
					DotEdge edge = dot.addEdge(nodeMap.get(fromNode), nodeMap.get(toNode),
							(int) relatedTypeCounts.get(fromNode, toNode).getMin() + ".." +
							(int) relatedTypeCounts.get(fromNode, toNode).getMax());
					
					edge.addMouseListener(new EdgeMouseListener(fromNode, toNode));
				}
			}
		}
		graph.changeDot(dot, true);
		
		graph.addMouseMotionListener(new GraphMouseMotionListener());
		
		return graph;
	}
	
	private class GraphMouseMotionListener implements MouseMotionListener {

		public void mouseDragged(MouseEvent e) {
			
		}

		public void mouseMoved(MouseEvent e) {
			if (currentStatisticsWindow != null && System.currentTimeMillis() - enterTime > 300 && exited) {
				currentStatisticsWindow.setVisible(false);
			}
		}
		
	}
	
	private class EdgeMouseListener implements MouseListener {
		
		DescriptiveStatistics edgeStatistics;
		JWindow statisticsWindow = new JWindow();

		EdgeMouseListener(String fromNode, String toNode) {
			edgeStatistics = relatedTypeCounts.get(fromNode, toNode);
			
			statisticsWindow.setLayout(new BoxLayout(statisticsWindow.getContentPane(), BoxLayout.PAGE_AXIS));
			
			addStatisticsLabel("Instances of type: " + fromNode);
			addStatisticsLabel("Relate to instances of: " + toNode);
			addStatisticsLabel(" ");
			addStatisticsLabel("Minimum: " + (int) edgeStatistics.getMin());
			addStatisticsLabel("Maximum: " + (int) edgeStatistics.getMax());
			addStatisticsLabel("Mean: " + edgeStatistics.getMean());
			addStatisticsLabel("Median: " + edgeStatistics.getPercentile(50));
			addStatisticsLabel("Sum: " + (int) edgeStatistics.getSum());
			
			statisticsWindow.setSize(300, 155);
		}
		
		private void addStatisticsLabel(String text) {
			JLabel label = new JLabel(text, SwingConstants.LEFT);
			label.setAlignmentX(Component.LEFT_ALIGNMENT);
			label.setFont(new Font("Dialog", Font.BOLD, 14));
			
			statisticsWindow.add(label);
		}

		public void mouseClicked(MouseEvent e) {
			
		}

		public void mousePressed(MouseEvent e) {
			
		}

		public void mouseReleased(MouseEvent e) {
			
		}

		public void mouseEntered(MouseEvent e) {
			if (currentStatisticsWindow != null) {
				currentStatisticsWindow.setVisible(false);
			}
		    
			statisticsWindow.setLocation(e.getXOnScreen()+10, e.getYOnScreen()+10);
			statisticsWindow.setVisible(true);
			
			enterTime = System.currentTimeMillis();
			currentStatisticsWindow = statisticsWindow;
			
			exited = false;
		}

		public void mouseExited(MouseEvent e) {
			if (System.currentTimeMillis() - enterTime > 300) {
				statisticsWindow.setVisible(false);
			}
			exited = true;
		}
		
	}

}
