package org.processmining.miminer.visualisation;

import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.util.HashMap;
import java.util.Map;

import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JWindow;
import javax.swing.KeyStroke;
import javax.swing.SwingConstants;

import org.apache.commons.lang3.time.DurationFormatUtils;
import org.apache.commons.math3.stat.descriptive.DescriptiveStatistics;
import org.processmining.miminer.models.ArtifactTypeModel;
import org.processmining.miminer.models.ArtifactTypeModel.ArtifactState;
import org.processmining.miminer.models.ArtifactTypeModel.ArtifactTransition;
import org.processmining.plugins.graphviz.dot.Dot;
import org.processmining.plugins.graphviz.dot.DotEdge;
import org.processmining.plugins.graphviz.dot.DotElement;
import org.processmining.plugins.graphviz.dot.DotNode;
import org.processmining.plugins.graphviz.visualisation.DotPanel;
import org.processmining.plugins.graphviz.visualisation.listeners.DotElementSelectionListener;

import com.kitfox.svg.SVGDiagram;
import com.kitfox.svg.SVGElement;

public class ArtifactTypeModelVisualisation extends DotPanel {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	
	private VisualisationConnections connections;
	private ArtifactTypeModel artifactTypeModel;
	private JWindow currentStatisticsWindow;
	private long enterTime;
	private boolean exited = true;

	public ArtifactTypeModelVisualisation(Dot dot, VisualisationConnections connections, ArtifactTypeModel artifactTypeModel) {
		super(dot);
		
		this.connections = connections;
		this.artifactTypeModel = artifactTypeModel;
		setBorder(BorderFactory.createEmptyBorder(2, 2, 2, 2));
		setBackground(Color.white);
		
		// Populate graph
		Map<String,DotNode> nodeMap = new HashMap<>();
		for (ArtifactState state : artifactTypeModel.artifactStates.elementSet()) {
			DotNode node = dot.addNode(state.stateName + "\n" + artifactTypeModel.artifactStates.count(state));
			nodeMap.put(state.stateName, node);
			
			node.addSelectionListener(new NodeSelectionListener(this, state));
			node.setSelectable(true);
			node.addMouseListener(new NodeMouseListener(state));
		}
		for (ArtifactTransition transition : artifactTypeModel.artifactTransitions.elementSet()) {
			DotEdge edge = dot.addEdge(nodeMap.get(transition.fromState), nodeMap.get(transition.toState),
					artifactTypeModel.artifactTransitions.count(transition) +"");
			
			edge.addSelectionListener(new EdgeSelectionListener(this, transition));
			edge.setSelectable(true);
		}
		changeDot(dot, true);
		
		// request keyboard focus when mouse is on graph, fix this shit for all keybindings
		getInputMap(JComponent.WHEN_FOCUSED).put(KeyStroke.getKeyStroke(KeyEvent.VK_EQUALS, InputEvent.CTRL_MASK), "zoomIn");
		getInputMap(JComponent.WHEN_FOCUSED).put(KeyStroke.getKeyStroke(KeyEvent.VK_PLUS, InputEvent.CTRL_MASK), "zoomIn");
		getInputMap(JComponent.WHEN_FOCUSED).put(KeyStroke.getKeyStroke(KeyEvent.VK_MINUS, InputEvent.CTRL_MASK), "zoomOut");
		getInputMap(JComponent.WHEN_FOCUSED).put(KeyStroke.getKeyStroke(KeyEvent.VK_0, InputEvent.CTRL_MASK), "viewReset");
		getInputMap(JComponent.WHEN_FOCUSED).put(KeyStroke.getKeyStroke(KeyEvent.VK_I, InputEvent.CTRL_MASK), "saveAs");
		getInputMap(JComponent.WHEN_FOCUSED).put(KeyStroke.getKeyStroke(KeyEvent.VK_D, InputEvent.CTRL_MASK), "changeGraphDirection");
		getInputMap(JComponent.WHEN_FOCUSED).put(KeyStroke.getKeyStroke(KeyEvent.VK_Q, InputEvent.CTRL_MASK), "increaseAspectRatio");
		getInputMap(JComponent.WHEN_FOCUSED).put(KeyStroke.getKeyStroke(KeyEvent.VK_W, InputEvent.CTRL_MASK), "decreaseAspectRatio");
		addMouseListener(new GraphMouseListener(this));
		
		addMouseMotionListener(new GraphMouseMotionListener());
	}
	
	private String formatDuration(long duration) {
		return DurationFormatUtils.formatDuration(duration,"d 'days' HH:mm:ss",true);
	}

	private class NodeSelectionListener implements DotElementSelectionListener {

		DotPanel graph;
		ArtifactState state;
		
		public NodeSelectionListener(DotPanel graph, ArtifactState state) {
			this.graph = graph;
			this.state = state;
		}

		public void selected(DotElement element, SVGDiagram image) {
			SVGElement node = graph.getSVGElementOf(graph.getSVG(), element).getChild(1);
			graph.setCSSAttributeOf(node, "stroke", "red");
			graph.setImage(graph.getSVG(), false);
			
			for (String type : connections.graphs.keySet()) {
				if (!type.equals(artifactTypeModel.artifactType)) {
					connections.graphs.get(type).changeDot(connections.graphs.get(type).getDot(), false);
				}
			}
			
			connections.selectedState = state;
			connections.selectedModel = artifactTypeModel;
			connections.statisticsWindows.clear();
			
			// TODO: remove improper selection behaviour below
			System.out.println(state + " " + artifactTypeModel.artifactStates.count(state));
			DescriptiveStatistics periodStatistics = artifactTypeModel.statePeriods.get(state);
			if (periodStatistics != null) {
				System.out.println(periodStatistics.getN() + " periods: " + formatDuration((long) periodStatistics.getMean()));
			}
			
//			for (String type : artifactTypeModel.relatedStateCounts.row(state).keySet()) {
//				System.out.println(type + artifactTypeModel.relatedStateCounts.get(state, type));
//			}
		}

		public void deselected(DotElement element, SVGDiagram image) {
			SVGElement node = graph.getSVGElementOf(graph.getSVG(), element).getChild(1);
			graph.setCSSAttributeOf(node, "stroke", "black");
			graph.setImage(graph.getSVG(), false);
			
			connections.selectedState = null;
			connections.selectedModel = null;
		}
		
	}
	
	private class EdgeSelectionListener implements DotElementSelectionListener {

		DotPanel graph;
		ArtifactTransition transition;
		
		public EdgeSelectionListener(DotPanel graph, ArtifactTransition transition) {
			this.graph = graph;
			this.transition = transition;
		}

		public void selected(DotElement element, SVGDiagram image) {
			SVGElement edge = graph.getSVGElementOf(graph.getSVG(), element).getChild(1);
			graph.setCSSAttributeOf(edge, "stroke", "red");
			graph.setImage(graph.getSVG(), false);
			
			for (String type : connections.graphs.keySet()) {
				if (!type.equals(artifactTypeModel.artifactType)) {
					connections.graphs.get(type).changeDot(connections.graphs.get(type).getDot(), false);
				}
			}
			
			connections.selectedTransition = transition;
			connections.selectedModel = artifactTypeModel;
			connections.statisticsWindows.clear();
			
			// TODO: remove improper selection behaviour below
//			System.out.println("test:" + artifactTypeModel.transitionRelatedTransitionCounts.row(transition));
		}

		public void deselected(DotElement element, SVGDiagram image) {
			SVGElement edge = graph.getSVGElementOf(graph.getSVG(), element).getChild(1);
			graph.setCSSAttributeOf(edge, "stroke", "black");
			graph.setImage(graph.getSVG(), false);
			
			connections.selectedTransition = null;
			connections.selectedModel = null;
		}
		
	}
	
	private class NodeMouseListener implements MouseListener {
		
		ArtifactState state;
		JWindow statisticsWindow = new JWindow();
		
		NodeMouseListener(ArtifactState state) {
			this.state = state;
			
			statisticsWindow.setLayout(new BoxLayout(statisticsWindow.getContentPane(), BoxLayout.PAGE_AXIS));
			
			addStatisticsLabel("Instances of type: " + state.artifactType);
			addStatisticsLabel("Having state: " + state.stateName);
			addStatisticsLabel(" ");
			if (artifactTypeModel.statePeriods.containsKey(state)) {
				addStatisticsLabel("Periods: " + artifactTypeModel.statePeriods.get(state).getN());
				addStatisticsLabel("Minimum duration: " + formatDuration((long) artifactTypeModel.statePeriods.get(state).getMin()));
				addStatisticsLabel("Maximum duration: " + formatDuration((long) artifactTypeModel.statePeriods.get(state).getMax()));
				addStatisticsLabel("Mean duration: " + formatDuration((long) artifactTypeModel.statePeriods.get(state).getMean()));
				addStatisticsLabel("Median duration: " + formatDuration((long) artifactTypeModel.statePeriods.get(state).getPercentile(50)));
			}
			else {
				
			}
			
			statisticsWindow.setSize(300, 155);
		}
		
		private void addStatisticsLabel(String text) {
			JLabel label = new JLabel(text, SwingConstants.LEFT);
			label.setAlignmentX(Component.LEFT_ALIGNMENT);
			label.setFont(new Font("Dialog", Font.BOLD, 14));
			
			statisticsWindow.add(label);
		}
		
		private void addCurrentStatisticsLabel(String text) {
			JLabel label = new JLabel(text, SwingConstants.LEFT);
			label.setAlignmentX(Component.LEFT_ALIGNMENT);
			label.setFont(new Font("Dialog", Font.BOLD, 14));
			
			currentStatisticsWindow.add(label);
		}

		public void mouseClicked(MouseEvent e) {
			
		}

		public void mousePressed(MouseEvent e) {
			
		}

		public void mouseReleased(MouseEvent e) {
			
		}

		public void mouseEntered(MouseEvent e) {
			if (currentStatisticsWindow != null) {
				currentStatisticsWindow.setVisible(false);
			}
		    
			if (connections.selectedState == null) {
				currentStatisticsWindow = statisticsWindow;
			}
			else {
				currentStatisticsWindow = connections.statisticsWindows.get(state);
				
				if (currentStatisticsWindow == null) {
					currentStatisticsWindow = new JWindow();
					
					currentStatisticsWindow.setLayout(new BoxLayout(currentStatisticsWindow.getContentPane(), BoxLayout.PAGE_AXIS));
					
					// TODO: add content based on selected state
					addCurrentStatisticsLabel("If an instance of type: " + connections.selectedState.artifactType);
					addCurrentStatisticsLabel("Is in state: " + connections.selectedState.stateName);
					addCurrentStatisticsLabel(" ");
					addCurrentStatisticsLabel("Time spent with at least one");
					addCurrentStatisticsLabel("instance in: " + state.stateName);
					addCurrentStatisticsLabel(" ");
					
					if (connections.selectedModel.relatedStatePeriods.contains(connections.selectedState, state)) {
						addCurrentStatisticsLabel("Average time: " + 
								formatDuration((long) connections.selectedModel.relatedStatePeriods.get(
										connections.selectedState, state).getMean()));
					}
					
					currentStatisticsWindow.setSize(300, 155);
				}
			}
			
			currentStatisticsWindow.setLocation(e.getXOnScreen()+10, e.getYOnScreen()+10);
			currentStatisticsWindow.setVisible(true);
			
			enterTime = System.currentTimeMillis();
			exited = false;
		}

		public void mouseExited(MouseEvent e) {
			if (System.currentTimeMillis() - enterTime > 300) {
				statisticsWindow.setVisible(false);
			}
			exited = true;
		}
		
	}
	
	private class GraphMouseMotionListener implements MouseMotionListener {

		public void mouseDragged(MouseEvent e) {
			
		}

		public void mouseMoved(MouseEvent e) {
			if (currentStatisticsWindow != null && System.currentTimeMillis() - enterTime > 300 && exited) {
				currentStatisticsWindow.setVisible(false);
			}
		}
		
	}
	
	private class GraphMouseListener implements MouseListener {

		DotPanel graphPanel;
		
		GraphMouseListener(DotPanel graphPanel) {
			this.graphPanel = graphPanel;
		}
		
		public void mouseClicked(MouseEvent e) {
			
		}

		public void mousePressed(MouseEvent e) {
			graphPanel.requestFocusInWindow();
		}

		public void mouseReleased(MouseEvent e) {
			
		}

		public void mouseEntered(MouseEvent e) {
			graphPanel.requestFocusInWindow();
		}

		public void mouseExited(MouseEvent e) {
			if (currentStatisticsWindow != null) {
				currentStatisticsWindow.setVisible(false);
			}
		}
		
	}
}
