package org.processmining.miminer.visualisation;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.swing.BorderFactory;
import javax.swing.DefaultCellEditor;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JPanel;
import javax.swing.JSplitPane;
import javax.swing.JTextField;
import javax.swing.ListSelectionModel;
import javax.swing.RowFilter;
import javax.swing.RowFilter.ComparisonType;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableColumn;
import javax.swing.table.TableModel;
import javax.swing.table.TableRowSorter;

import org.apache.commons.lang3.time.DurationFormatUtils;
import org.apache.commons.math3.stat.descriptive.DescriptiveStatistics;
import org.processmining.framework.util.ui.widgets.ProMTable;
import org.processmining.framework.util.ui.widgets.ProMTextField;
import org.processmining.miminer.MIMinerResults;
import org.processmining.miminer.models.ArtifactTypeModel;
import org.processmining.miminer.models.ArtifactTypeModel.ArtifactState;
import org.processmining.miminer.models.ArtifactTypeModel.ArtifactTransition;

import com.fluxicon.slickerbox.factory.SlickerFactory;
import com.google.common.collect.HashBasedTable;
import com.google.common.collect.HashMultimap;
import com.google.common.collect.HashMultiset;
import com.google.common.collect.Multiset;
import com.google.common.collect.SetMultimap;
import com.google.common.collect.Table;

public class RelationsPanel extends JSplitPane {

	private static final long serialVersionUID = 3468816272629744389L;
	
	public JPanel panel;
	public ProMTable table;
	
	private List<Integer> relationTypePerRow = new ArrayList<>();
	private List<String> conditionTypesPerRow = new ArrayList<>();
	private List<String> conditionNamesPerRow = new ArrayList<>();
	private SetMultimap<Integer, String> consequenceTypesPerRow = HashMultimap.create();
	private List<String> consequenceNamesPerRow = new ArrayList<>();
	private Map<Integer, Double> minValuesPerColumn = new HashMap<>();
	private TableRowSorter<TableModel> sorter;
	
	private Dimension singleSize, doubleSize, tripleSize;
	private String[] relationTypes = {
			"Transition-State co-occurrences",
			"Transition-Transition co-occurrences",
			"Transition-Transition multisets",
			"Transition markings",
			"State-Transition co-occurrences",
			"State-Transition multisets",
			"State-Transition markings",
			"State-State periods",
			"State-Marking periods",
			"State-Type marking periods"
			};
	private JComboBox switchRelationTypeBox = SlickerFactory.instance().createComboBox(relationTypes);
	private JCheckBox toggleTypeNameBox = SlickerFactory.instance().createCheckBox("Show type names", false);
	private JComboBox conditionArtifactTypeBox;
	private ProMTextField conditionFilter = new ProMTextField("Filter conditions by name");
	private JComboBox consequenceArtifactTypeBox;
	private ProMTextField consequenceFilter = new ProMTextField("Filter consequences by name");
	private ProMTextField minConfidence = new ProMTextField("Minimum Confidence");
	private ProMTextField minLift = new ProMTextField("Minimum Lift");
	
	public RelationsPanel(MIMinerResults results, VisualisationConnections connections, int width) {
		super(JSplitPane.VERTICAL_SPLIT);
		
		this.table = new ProMTable(relationsTableModel);
		this.panel = new JPanel(new GridBagLayout());
		
		setBackground(Color.LIGHT_GRAY);
		setBottomComponent(table);
		setTopComponent(panel);
		setResizeWeight(0.0);
		setDividerSize(0);
		setBorder(BorderFactory.createEmptyBorder(1,1,1,1));
		
		conditionArtifactTypeBox = SlickerFactory.instance().createComboBox(results.configuration.types.toArray(new String[0]));
		
		// TODO: add <all types> entry
		List<String> otherTypes = new ArrayList<>();
		otherTypes.add("<all types>");
		otherTypes.addAll(results.configuration.types);
		consequenceArtifactTypeBox = SlickerFactory.instance().createComboBox(otherTypes.toArray(new String[0]));
		createTable(width);
		
		sorter = new TableRowSorter<TableModel>(table.getTable().getModel());
		table.setRowSorter(sorter);
		
		// TODO: add content
		addRelations(results);
		sorter.setRowFilter(createCombinedFilter());
		
		// add filtering listeners
		switchRelationTypeBox.addActionListener(new TypeActionListener());
		toggleTypeNameBox.addActionListener(new ToggleTypeNameActionListener());
		conditionArtifactTypeBox.addActionListener(new TypeActionListener());
		conditionFilter.getDocument().addDocumentListener(new FilterDocumentListener());
		consequenceArtifactTypeBox.addActionListener(new TypeActionListener());
		consequenceFilter.getDocument().addDocumentListener(new FilterDocumentListener());
		minConfidence.getDocument().addDocumentListener(new MinMetricDocumentListener(minConfidence, 3));
		minLift.getDocument().addDocumentListener(new MinMetricDocumentListener(minLift, 4));
		
		connections.relationsPane = this;
	}
	
	private void addRelations(MIMinerResults results) {
		for (String type : results.artifactTypeModels.keySet()) {
			ArtifactTypeModel model = results.artifactTypeModels.get(type);
			
			// Calculate confidence = ((X & Y) / X) and lift = confidence / (Y / Total) for transition relations
			Integer sumTransitionCounts = 0;
			Multiset<ArtifactState> consequenceSumTransitionPostStateCounts = HashMultiset.create();
			Multiset<ArtifactTransition> consequenceSumTransitionPostTransitionCounts = HashMultiset.create();
			Multiset<Multiset<ArtifactTransition>> consequenceSumTransitionPostMultisets = HashMultiset.create();
			Multiset<Multiset<ArtifactTransition>> consequenceSumTransitionPostMarkings = HashMultiset.create();
			for (ArtifactTransition transition : model.artifactTransitions.elementSet()) {
				sumTransitionCounts += model.artifactTransitions.count(transition);
				
				for (ArtifactState state : model.transitionPostStateCounts.row(transition).keySet()) {
					consequenceSumTransitionPostStateCounts.add(state, model.transitionPostStateCounts.get(transition, state));
				}
				
				for (ArtifactTransition otherTransition : model.transitionPostTransitionCounts.row(transition).keySet()) {
					consequenceSumTransitionPostTransitionCounts.add(otherTransition, model.transitionPostTransitionCounts.get(
							transition, otherTransition));
				}
				
				for (Multiset<ArtifactTransition> multiset : model.transitionPostMultisets.row(transition).keySet()) {
					if (multiset.size() == 1) continue;
					consequenceSumTransitionPostMultisets.add(multiset, model.transitionPostMultisets.get(transition, multiset));
				}
				
				for (Multiset<ArtifactTransition> marking : model.transitionPostMarkings.row(transition).keySet()) {
					consequenceSumTransitionPostMarkings.add(marking, model.transitionPostMarkings.get(transition, marking));
				}
			}
			
			// Add transition relations
			for (ArtifactTransition transition : model.artifactTransitions.elementSet()) {
				// "Transition-State co-occurrences"
				for (ArtifactState state : model.transitionPostStateCounts.row(transition).keySet()) {
					String condition = transition.toString();
					String consequence = state.toString();
					Integer value = model.transitionPostStateCounts.get(transition, state);
					Double confidence = ((double) value / model.artifactTransitions.count(transition));
					Double lift = confidence / ((double) consequenceSumTransitionPostStateCounts.count(state) / sumTransitionCounts);
					
					relationsTableModel.addRow(new Object[] { condition, consequence, value.toString(), confidence, lift });
					relationTypePerRow.add(0);
					conditionTypesPerRow.add(type);
					consequenceTypesPerRow.put(relationsTableModel.getRowCount() - 1, state.artifactType);
				}
				
				// "Transition-Transition co-occurrences"
				for (ArtifactTransition otherTransition : model.transitionPostTransitionCounts.row(transition).keySet()) {
					String condition = transition.toString();
					String consequence = otherTransition.toString();
					Integer value = model.transitionPostTransitionCounts.get(transition, otherTransition);
					Double confidence = ((double) value / model.artifactTransitions.count(transition));
					Double lift = confidence / ((double) consequenceSumTransitionPostTransitionCounts.count(
							otherTransition) / sumTransitionCounts);
					
					relationsTableModel.addRow(new Object[] { condition, consequence, value.toString(), confidence, lift });
					relationTypePerRow.add(1);
					conditionTypesPerRow.add(type);
					consequenceTypesPerRow.put(relationsTableModel.getRowCount() - 1, otherTransition.artifactType);
				}
				
				// "Transition-Transition multisets"
				for (Multiset<ArtifactTransition> multiset : model.transitionPostMultisets.row(transition).keySet()) {
					if (multiset.size() == 1) continue;
					
					String condition = transition.toString();
					String consequence = multiset.toString();
					Integer value = model.transitionPostMultisets.get(transition, multiset);
					Double confidence = ((double) value / model.artifactTransitions.count(transition));
					Double lift = confidence / ((double) consequenceSumTransitionPostMultisets.count(
							multiset) / sumTransitionCounts);
					
					relationsTableModel.addRow(new Object[] { condition, consequence, value.toString(), confidence, lift });
					relationTypePerRow.add(2);
					conditionTypesPerRow.add(type);
					for (ArtifactTransition otherTransition : multiset.elementSet()) {
						consequenceTypesPerRow.put(relationsTableModel.getRowCount() - 1, otherTransition.artifactType);
					}
				}
				
				// "Transition markings"
				for (Multiset<ArtifactTransition> marking : model.transitionPostMarkings.row(transition).keySet()) {
					String condition = transition.toString();
					String consequence = marking.toString();
					Integer value = model.transitionPostMarkings.get(transition, marking);
					Double confidence = ((double) value / model.artifactTransitions.count(transition));
					Double lift = confidence / ((double) consequenceSumTransitionPostMarkings.count(
							marking) / sumTransitionCounts);
					
					relationsTableModel.addRow(new Object[] { condition, consequence, value.toString(), confidence, lift });
					relationTypePerRow.add(3);
					conditionTypesPerRow.add(type);
					for (ArtifactTransition otherTransition : marking.elementSet()) {
						consequenceTypesPerRow.put(relationsTableModel.getRowCount() - 1, otherTransition.artifactType);
					}
				}
			}
			
			// Calculate confidence = ((X & Y) / X) and lift = confidence / (Y / Total) for state-transition relations
			Table<ArtifactState, ArtifactState, Integer> conditionSumRelatedTransitionCounts = HashBasedTable.create(); // X
			Multiset<ArtifactTransition> consequenceSumRelatedTransitionCounts = HashMultiset.create(); // Y
			Multiset<ArtifactState> consequenceStatesRelatedTransitionCounts = HashMultiset.create(); // Total
			Table<ArtifactState, Multiset<ArtifactState>, Integer> conditionSumrelatedTransitionMultisets = HashBasedTable.create();
			Multiset<Multiset<ArtifactTransition>> consequenceSumRelatedTransitionMultisets = HashMultiset.create();
			Multiset<Multiset<ArtifactState>> consequenceStatesRelatedTransitionMultisets = HashMultiset.create();
			Table<ArtifactState, Multiset<ArtifactState>, Integer> conditionSumrelatedTransitionMarkings = HashBasedTable.create();
			Multiset<Multiset<ArtifactTransition>> consequenceSumRelatedTransitionMarkings = HashMultiset.create();
			Multiset<Multiset<ArtifactState>> consequenceStatesRelatedTransitionMarkings = HashMultiset.create();
			for (ArtifactState state : model.artifactStates.elementSet()) {
				for (ArtifactTransition transition : model.relatedTransitionCounts.row(state).keySet()) {
					ArtifactState otherState = model.getState(transition.artifactType, transition.fromState);
					Integer count = conditionSumRelatedTransitionCounts.get(state, otherState);
					if (count == null) count = 0;
					conditionSumRelatedTransitionCounts.put(state, otherState, count + model.relatedTransitionCounts.get(state, transition));
					consequenceSumRelatedTransitionCounts.add(transition, model.relatedTransitionCounts.get(state, transition));
					consequenceStatesRelatedTransitionCounts.add(otherState, model.relatedTransitionCounts.get(state, transition));
				}
				
				for (Multiset<ArtifactTransition> multiset : model.relatedTransitionMultisets.row(state).keySet()) {
					Multiset<ArtifactState> stateMultiset = HashMultiset.create();
					for (ArtifactTransition transition : multiset.elementSet()) {
						stateMultiset.add(model.getState(transition.artifactType, transition.fromState), multiset.count(transition));
					}
					Integer count = conditionSumrelatedTransitionMultisets.get(state, stateMultiset);
					if (count == null) count = 0;
					conditionSumrelatedTransitionMultisets.put(state, stateMultiset,
							count + model.relatedTransitionMultisets.get(state, multiset));
					consequenceSumRelatedTransitionMultisets.add(multiset, model.relatedTransitionMultisets.get(state, multiset));
					consequenceStatesRelatedTransitionMultisets.add(stateMultiset, model.relatedTransitionMultisets.get(state, multiset));
				}
				
				for (Multiset<ArtifactTransition> marking : model.relatedTransitionMarkings.row(state).keySet()) {
					Multiset<ArtifactState> stateMarking = HashMultiset.create();
					for (ArtifactTransition transition : marking.elementSet()) {
						stateMarking.add(model.getState(transition.artifactType, transition.fromState), marking.count(transition));
					}
					Integer count = conditionSumrelatedTransitionMarkings.get(state, stateMarking);
					if (count == null) count = 0;
					conditionSumrelatedTransitionMarkings.put(state, stateMarking,
							count + model.relatedTransitionMarkings.get(state, marking));
					consequenceSumRelatedTransitionMarkings.add(marking, model.relatedTransitionMarkings.get(state, marking));
					consequenceStatesRelatedTransitionMarkings.add(stateMarking, model.relatedTransitionMarkings.get(state, marking));
				}
			}
			
			// Add state-transition relations
			for (ArtifactState state : model.artifactStates.elementSet()) {
				// "State-Transition co-occurrences"
				for (ArtifactTransition transition : model.relatedTransitionCounts.row(state).keySet()) {
					ArtifactState otherState = model.getState(transition.artifactType, transition.fromState);
					
					String condition = state.toString();
					String consequence = transition.toString();
					Integer value = model.relatedTransitionCounts.get(state, transition);
					Double confidence = ((double) value / conditionSumRelatedTransitionCounts.get(state, otherState));
					Double lift = confidence / ((double) consequenceSumRelatedTransitionCounts.count(transition) / 
							consequenceStatesRelatedTransitionCounts.count(otherState));
					
					relationsTableModel.addRow(new Object[] { condition, consequence, value.toString(), confidence, lift });
					relationTypePerRow.add(4);
					conditionTypesPerRow.add(type);
					consequenceTypesPerRow.put(relationsTableModel.getRowCount() - 1, transition.artifactType);
				}
				
				// "State-Transition multisets"
				for (Multiset<ArtifactTransition> multiset : model.relatedTransitionMultisets.row(state).keySet()) {
					Multiset<ArtifactState> stateMultiset = HashMultiset.create();
					for (ArtifactTransition transition : multiset.elementSet()) {
						stateMultiset.add(model.getState(transition.artifactType, transition.fromState), multiset.count(transition));
					}
					
					String condition = state.toString();
					String consequence = multiset.toString();
					Integer value = model.relatedTransitionMultisets.get(state, multiset);
					Double confidence = ((double) value / conditionSumrelatedTransitionMultisets.get(state, stateMultiset));
					Double lift = confidence / ((double) consequenceSumRelatedTransitionMultisets.count(multiset) / 
							consequenceStatesRelatedTransitionMultisets.count(stateMultiset));
					
					relationsTableModel.addRow(new Object[] { condition, consequence, value.toString(), confidence, lift });
					relationTypePerRow.add(5);
					conditionTypesPerRow.add(type);
					for (ArtifactTransition otherTransition : multiset.elementSet()) {
						consequenceTypesPerRow.put(relationsTableModel.getRowCount() - 1, otherTransition.artifactType);
					}
				}
				
				// "State-Transition markings"
				for (Multiset<ArtifactTransition> marking : model.relatedTransitionMarkings.row(state).keySet()) {
					Multiset<ArtifactState> stateMultiset = HashMultiset.create();
					for (ArtifactTransition transition : marking.elementSet()) {
						stateMultiset.add(model.getState(transition.artifactType, transition.fromState), marking.count(transition));
					}
					
					String condition = state.toString();
					String consequence = marking.toString();
					Integer value = model.relatedTransitionMarkings.get(state, marking);
					Double confidence = ((double) value / conditionSumrelatedTransitionMarkings.get(state, stateMultiset));
					Double lift = confidence / ((double) consequenceSumRelatedTransitionMarkings.count(marking) / 
							consequenceStatesRelatedTransitionMarkings.count(stateMultiset));
					
					relationsTableModel.addRow(new Object[] { condition, consequence, value.toString(), confidence, lift });
					relationTypePerRow.add(6);
					conditionTypesPerRow.add(type);
					for (ArtifactTransition otherTransition : marking.elementSet()) {
						consequenceTypesPerRow.put(relationsTableModel.getRowCount() - 1, otherTransition.artifactType);
					}
				}
			}
			
			// Calculate confidence = ((X & Y) / X) and lift = confidence / (Y / Total) for state-state relations
			Double sumStateTime = 0d;
			Table<ArtifactState, ArtifactState, DescriptiveStatistics> sumRelatedStatePeriods = HashBasedTable.create(); // >= 1 token
			Map<ArtifactState, DescriptiveStatistics> conditionSumRelatedStatePeriods = new HashMap<>(); // X
			Map<ArtifactState, DescriptiveStatistics> consequenceSumRelatedStatePeriods = new HashMap<>(); // Y
			Map<Multiset<ArtifactState>, DescriptiveStatistics> consequenceSumRelatedMarkings = new HashMap<>();
			Table<String, Multiset<ArtifactState>, DescriptiveStatistics> consequenceSumRelatedTypeMarkings = HashBasedTable.create();
			for (ArtifactState state : model.artifactStates.elementSet()) {
				for (Multiset<ArtifactState> marking : model.relatedMarkings.row(state).keySet()) {
					Double value = model.relatedMarkings.get(state, marking).getSum();
					sumStateTime += value;
					DescriptiveStatistics stateStatistics = conditionSumRelatedStatePeriods.get(state);
					if (stateStatistics == null) {
						stateStatistics = new DescriptiveStatistics();
						conditionSumRelatedStatePeriods.put(state, stateStatistics);
					}
					stateStatistics.addValue(value);

					for (ArtifactState otherState : marking.elementSet()) {
						DescriptiveStatistics relatedStatistics = sumRelatedStatePeriods.get(state, otherState);
						if (relatedStatistics == null) {
							relatedStatistics = new DescriptiveStatistics();
							sumRelatedStatePeriods.put(state, otherState, relatedStatistics);
						}
						relatedStatistics.addValue(value);
						
						DescriptiveStatistics otherStatistics = consequenceSumRelatedStatePeriods.get(otherState);
						if (otherStatistics == null) {
							otherStatistics = new DescriptiveStatistics();
							consequenceSumRelatedStatePeriods.put(otherState, otherStatistics);
						}
						otherStatistics.addValue(value);
					}
					
					DescriptiveStatistics markingStatistics = consequenceSumRelatedMarkings.get(marking);
					if (markingStatistics == null) {
						markingStatistics = new DescriptiveStatistics();
						consequenceSumRelatedMarkings.put(marking, markingStatistics);
					}
					markingStatistics.addValue(value);
				}
				
				for (String relatedType : model.relatedTypeMarkings.keySet()) {
					Table<ArtifactState, Multiset<ArtifactState>, DescriptiveStatistics> typeMarkings = 
							model.relatedTypeMarkings.get(relatedType);
					
					for (Multiset<ArtifactState> typeMarking : typeMarkings.row(state).keySet()) {
						DescriptiveStatistics typeMarkingStatistics = typeMarkings.get(state, typeMarking);
						DescriptiveStatistics sumTypeMarkingStatistics = consequenceSumRelatedTypeMarkings.get(relatedType, typeMarking);
						if (sumTypeMarkingStatistics == null) {
							sumTypeMarkingStatistics = new DescriptiveStatistics();
							consequenceSumRelatedTypeMarkings.put(relatedType, typeMarking, sumTypeMarkingStatistics);
						}
						sumTypeMarkingStatistics.addValue(typeMarkingStatistics.getSum());
					}
				}
			}
			
			for (ArtifactState state : model.artifactStates.elementSet()) {
				// TODO: print conditionSumRelatedStatePeriods v.s. normal state periods: conclusion they're equal
//				System.out.println(state + ":" + conditionSumRelatedStatePeriods.get(state).getSum() + " v.s. " + model.statePeriods.get(state).getSum());
				
				// "State-State periods"
				for (ArtifactState otherState : model.relatedStatePeriods.row(state).keySet()) {
					DescriptiveStatistics statistics = sumRelatedStatePeriods.get(state, otherState);
					
					String condition = state.toString();
					String consequence = otherState.toString();
					long value = (long) statistics.getSum();
					Double confidence = ((double) value / conditionSumRelatedStatePeriods.get(state).getSum());
					Double lift = confidence / (consequenceSumRelatedStatePeriods.get(otherState).getSum() / sumStateTime);
					
					relationsTableModel.addRow(new Object[] { condition, consequence, formatDuration(value), confidence, lift });
					relationTypePerRow.add(7);
					conditionTypesPerRow.add(type);
					consequenceTypesPerRow.put(relationsTableModel.getRowCount() - 1, otherState.artifactType);
				}
				
				// "State-Marking periods"
				for (Multiset<ArtifactState> marking : model.relatedMarkings.row(state).keySet()) {
					DescriptiveStatistics statistics = model.relatedMarkings.get(state, marking);
					
					String condition = state.toString();
					String consequence = marking.toString();
					long value = (long) statistics.getSum();
					Double confidence = ((double) value / conditionSumRelatedStatePeriods.get(state).getSum());
					Double lift = confidence / (consequenceSumRelatedMarkings.get(marking).getSum() / sumStateTime);
					lift = null;
					
					relationsTableModel.addRow(new Object[] { condition, consequence, formatDuration(value), confidence, lift });
					relationTypePerRow.add(8);
					conditionTypesPerRow.add(type);
					for (ArtifactState otherState : marking.elementSet()) {
						consequenceTypesPerRow.put(relationsTableModel.getRowCount() - 1, otherState.artifactType);
					}
				}
				
				// "State-Type marking periods"
				for (String relatedType : model.relatedTypeMarkings.keySet()) {
					Table<ArtifactState, Multiset<ArtifactState>, DescriptiveStatistics> typeMarkings = 
							model.relatedTypeMarkings.get(relatedType);
					
					for (Multiset<ArtifactState> typeMarking : typeMarkings.row(state).keySet()) {
						DescriptiveStatistics statistics = typeMarkings.get(state, typeMarking);
						
						String condition = state.toString();
						String consequence = typeMarking.toString();
						long value = (long) statistics.getSum();
						Double confidence = ((double) value / conditionSumRelatedStatePeriods.get(state).getSum());
						Double lift = confidence / (consequenceSumRelatedTypeMarkings.get(relatedType, typeMarking).getSum() / sumStateTime);
						
						relationsTableModel.addRow(new Object[] { condition, consequence, formatDuration(value), confidence, lift });
						relationTypePerRow.add(9);
						conditionTypesPerRow.add(type);
						consequenceTypesPerRow.put(relationsTableModel.getRowCount() - 1, relatedType);
					}
				}
			}
		}
	}

	private void createTable(int width) {
		int columnCount = table.getColumnModel().getColumnCount();
		singleSize = new Dimension((width - 0) / 12,23);
		doubleSize = new Dimension((width - 0) * 2 / 12,23);
		switchRelationTypeBox.setMinimumSize(doubleSize);
		toggleTypeNameBox.setMinimumSize(singleSize);
		conditionArtifactTypeBox.setMinimumSize(singleSize);
		conditionFilter.setMinimumSize(doubleSize);
		consequenceArtifactTypeBox.setMinimumSize(singleSize);
		consequenceFilter.setMinimumSize(doubleSize);
		minConfidence.setMinimumSize(singleSize);
		minLift.setMinimumSize(singleSize);
		
		GridBagConstraints constraints = new GridBagConstraints();
		constraints.weightx = 2;
		constraints.gridwidth = 2;
		panel.add(switchRelationTypeBox, constraints);
		constraints.weightx = 1;
		constraints.gridwidth = 1;
		panel.add(toggleTypeNameBox, constraints);
		panel.add(conditionArtifactTypeBox, constraints);
		constraints.weightx = 2;
		constraints.gridwidth = 2;
		panel.add(conditionFilter, constraints);
		constraints.weightx = 1;
		constraints.gridwidth = 1;
		panel.add(consequenceArtifactTypeBox, constraints);
		constraints.weightx = 2;
		constraints.gridwidth = 2;
		panel.add(consequenceFilter, constraints);
		constraints.weightx = 1;
		constraints.gridwidth = 1;
		panel.add(minConfidence, constraints);
		panel.add(minLift, constraints);
		
		panel.setPreferredSize(new Dimension(width, 23));
		panel.setMinimumSize(new Dimension(width, 23));
		panel.setBackground(Color.LIGHT_GRAY);
		
//		table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		table.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
		// TODO: add selection logic
//		table.getSelectionModel().addListSelectionListener(new RowListener());
		table.setMinimumSize(new Dimension(200, 40));
		table.setPreferredSize(new Dimension(1000, 90));
		table.getTableHeader().setPreferredSize(new Dimension(1000,20));
		table.getTableHeader().setReorderingAllowed(false);
		
		relationsTableModel.addColumn("Condition");
		relationsTableModel.addColumn("Consequence");
		relationsTableModel.addColumn("Value");
		relationsTableModel.addColumn("Confidence");
		relationsTableModel.addColumn("Lift");
		
		Enumeration<TableColumn> columns = table.getColumnModel().getColumns();
		while (columns.hasMoreElements()) {
			final TableColumn column = columns.nextElement();
			column.setPreferredWidth(width / (columnCount + 4));
			column.setCellEditor(new DefaultCellEditor(new JTextField()) {

				protected void fireEditingStopped() {
					this.cancelCellEditing();
					super.fireEditingStopped();
				}

				protected void fireEditingCanceled() {
					super.fireEditingCanceled();
				}

			});
		}
		table.getColumnModel().getColumn(0).setPreferredWidth(3*(width / (columnCount + 4)));
		table.getColumnModel().getColumn(1).setPreferredWidth(3*(width / (columnCount + 4)));
	}
	
	DefaultTableModel relationsTableModel = new DefaultTableModel() {
		public Class getColumnClass(int column) {
			switch (column) {
				case 0: case 1: case 2:
					return String.class;
				case 3: case 4:
					return Double.class;
				default:
					return Object.class;
			}
		}
	};
	
	private RowFilter<Object, Object> createCombinedFilter() {
		List<RowFilter<Object,Object>> filters = new ArrayList<>();
		filters.add(createRelationTypeFilter());
		filters.add(createConditionTypeFilter());
		filters.add(createConditionFilter());
		filters.add(createConsequenceTypeFilter());
		filters.add(createConsequenceFilter());
		
		RowFilter<Object, Object> valueFilter = createMinValueFilter();
		if (valueFilter != null) {
			filters.add(valueFilter);
		}
		
		return RowFilter.andFilter(filters);
	}
	
	private RowFilter<Object, Object> createRelationTypeFilter() {
		RowFilter<Object, Object> filter = new RowFilter<Object,Object>() {

			public boolean include(javax.swing.RowFilter.Entry<? extends Object, ? extends Object> entry) {
				return switchRelationTypeBox.getSelectedIndex() == relationTypePerRow.get((Integer) entry.getIdentifier());
			}
			
		};
		
		return filter;
	}
	
	private RowFilter<Object, Object> createConditionTypeFilter() {
		RowFilter<Object, Object> filter = new RowFilter<Object,Object>() {

			public boolean include(javax.swing.RowFilter.Entry<? extends Object, ? extends Object> entry) {
				return conditionArtifactTypeBox.getSelectedItem().equals(conditionTypesPerRow.get((Integer) entry.getIdentifier()));
			}
			
		};
		
		return filter;
	}
	
	private RowFilter<Object, Object> createConditionFilter() {
		RowFilter<Object, Object> filter = new RowFilter<Object,Object>() {

			public boolean include(javax.swing.RowFilter.Entry<? extends Object, ? extends Object> entry) {
				if (conditionFilter.getText().equals("Filter conditions by name")) return true;
				return entry.getStringValue(0).contains(conditionFilter.getText());
			}
			
		};
		
		return filter;
	}
	
	private RowFilter<Object, Object> createConsequenceTypeFilter() {
		RowFilter<Object, Object> filter = new RowFilter<Object,Object>() {

			public boolean include(javax.swing.RowFilter.Entry<? extends Object, ? extends Object> entry) {
				if (consequenceArtifactTypeBox.getSelectedIndex() == 0) return true;
				for (String type : consequenceTypesPerRow.get((Integer) entry.getIdentifier())) {
					if (consequenceArtifactTypeBox.getSelectedItem().equals(type)) return true;
				}
				return false;
			}
			
		};
		
		return filter;
	}
	
	private RowFilter<Object, Object> createConsequenceFilter() {
		RowFilter<Object, Object> filter = new RowFilter<Object,Object>() {

			public boolean include(javax.swing.RowFilter.Entry<? extends Object, ? extends Object> entry) {
				if (consequenceFilter.getText().equals("Filter consequences by name")) return true;
				return entry.getStringValue(1).contains(consequenceFilter.getText());
			}
			
		};
		
		return filter;
	}
	
	private RowFilter<Object, Object> createMinValueFilter() {
		RowFilter<Object, Object> filter = null;
		List<RowFilter<Object,Object>> filters = new ArrayList<>();
		for (Integer index : minValuesPerColumn.keySet()) {
			List<RowFilter<Object,Object>> greaterEqual = new ArrayList<>();
			greaterEqual.add(RowFilter.numberFilter(ComparisonType.AFTER, minValuesPerColumn.get(index), index));
			greaterEqual.add(RowFilter.numberFilter(ComparisonType.EQUAL, minValuesPerColumn.get(index), index));
			filters.add(RowFilter.orFilter(greaterEqual));
		}
		
		if (filters.size() > 0) {
			filter = RowFilter.andFilter(filters);
		}
		
		return filter;
	}
	
	private class TypeActionListener implements ActionListener {

		public void actionPerformed(ActionEvent e) {
			sorter.setRowFilter(createCombinedFilter());
		}
		
	}
	
	private class ToggleTypeNameActionListener implements ActionListener {

		public void actionPerformed(ActionEvent e) {
			for (int i = 0; i < relationsTableModel.getRowCount(); i++) {
				if (toggleTypeNameBox.isSelected()) {
					// TODO: use custom toString functions with boolean to show artifact type
//					relationsTableModel.setValueAt(conditionTypesPerRow.get(i) + ": " + conditionNamesPerRow.get(i), i, 0);
//					relationsTableModel.setValueAt(consequenceTypesPerRow.get(i) + ": " + consequenceNamesPerRow.get(i), i, 1);
				}
				else {
//					relationsTableModel.setValueAt(conditionNamesPerRow.get(i), i, 0);
//					relationsTableModel.setValueAt(consequenceNamesPerRow.get(i), i, 1);
				}
			}
			
			sorter.setRowFilter(createCombinedFilter());
		}
		
	}
	
	private class FilterDocumentListener implements DocumentListener {

		public void insertUpdate(DocumentEvent e) {
			sorter.setRowFilter(createCombinedFilter());
		}

		public void removeUpdate(DocumentEvent e) {
			sorter.setRowFilter(createCombinedFilter());
		}

		public void changedUpdate(DocumentEvent e) {
			sorter.setRowFilter(createCombinedFilter());
		}
		
	}
	
	private class MinMetricDocumentListener implements DocumentListener {
		
		private ProMTextField textField;
		private Integer columnIndex;
		
		public MinMetricDocumentListener(ProMTextField textField, int columnIndex) {
			this.textField = textField;
			this.columnIndex = columnIndex;
		}

		public void insertUpdate(DocumentEvent e) {
			updateFilter();
		}

		public void removeUpdate(DocumentEvent e) {
			updateFilter();
		}

		public void changedUpdate(DocumentEvent e) {
			updateFilter();
		}
		
		public void updateFilter() {
			try {
				Double minValue = Double.parseDouble(textField.getText());
				minValuesPerColumn.put(columnIndex, minValue);
			}
			catch (Exception e) {
				minValuesPerColumn.remove(columnIndex);
			}
			
			sorter.setRowFilter(createCombinedFilter());
		}
		
	}
	
	private String formatDuration(long duration) {
		return DurationFormatUtils.formatDuration(duration,"d 'days' HH:mm:ss.SSS",true);
	}
	
}
